#include <iostream>
#include <thread>
#include <vector>
#include <atomic>
#include <iterator>
#include <mutex>
#include <condition_variable>
#include <random>

#define LOG true

int THREADS = 4;
const int MATRIX_DIM = 5;//500;
const int TASKS = 10;//000;
std::vector<std::vector<int>> matrix(MATRIX_DIM);
std::vector<std::vector<std::recursive_mutex>> mutexes;

void add_number(std::vector<std::vector<int>>& matrix, int row, int column, int value) {
    std::lock_guard<std::recursive_mutex> l(mutexes[row][column]);
    matrix[row][column] += value;
}

void divide_by_number(std::vector<std::vector<int>>& matrix, int row, int column, int value) {
    std::lock_guard<std::recursive_mutex> l(mutexes[row][column]);
    matrix[row][column] = matrix[row][column] / value;
}

//there is a one dimension padding in the matrix so that we do not need to check boundaries
int average_with_neighbours(std::vector<std::vector<int>>& matrix, int row, int column) {
    std::lock(mutexes[row][column-1],mutexes[row][column],mutexes[row][column+1],mutexes[row-1][column],mutexes[row+1][column]);
    if (LOG) std::cout << "thread is averaging " << row << "," << column << " old_value: " << matrix[row][column] << std::endl;
    int v = (matrix[row][column-1] + matrix[row][column+1] + matrix[row-1][column] + matrix[row+1][column])/8;
    divide_by_number(matrix,row,column,2);
    add_number(matrix,row,column,v);
    std::this_thread::sleep_for(std::chrono::milliseconds(1));
    int result = matrix[row][column];
    if (LOG) std::cout << " new_value: " << result << std::endl;
    for (int i=-1;i<=1;i++)
        for (int j=-1; j<=1; j++) {
            if (i*j != 0) continue;
            mutexes[row+i][column+j].unlock();
        }
    return result;
}

struct TaskPool {
    const int MAXPOOL = 2;

    TaskPool(std::vector<std::vector<int>> *matrix) : matrix(matrix), buffer(MAXPOOL) {}
    std::vector<std::vector<int>>* matrix;

    int front = 0;
    int rear = 0;
    int count = 0;

    std::vector<std::pair<int,int>> buffer;

    std::mutex lock;

    std::condition_variable not_full;
    std::condition_variable not_empty;

    void add_task(int row, int column){
        std::unique_lock<std::mutex> l(lock);

        not_full.wait(l, [this](){return count != MAXPOOL; });

        buffer[rear] = std::pair<int,int>(row,column);
        rear = (rear + 1) % MAXPOOL;
        count++;

        not_empty.notify_one();
    }

    std::tuple<int,int,int> execute_task(){
        std::unique_lock<std::mutex> l(lock);

        not_empty.wait(l, [this](){return count != 0; });

        std::pair<int,int> square = buffer[front];
        front = (front + 1) % MAXPOOL;
        count--;

        not_full.notify_one();
        l.unlock();
        int result = average_with_neighbours(*matrix, square.first, square.second);

        return std::tuple<int,int,int>(result,square.first,square.second);
    }
};

void producer(int id, TaskPool& tp) {
    std::default_random_engine generator;
    std::uniform_int_distribution<int> distribution(1,MATRIX_DIM-2);

    for(int i = 0; i < TASKS; ++i){
        int r = distribution(generator);
        int c = distribution(generator);
        tp.add_task(r,c);
        if (LOG) std::cout << "Producer " << id << " added task [" << r << "," << c << "] " << std::endl;
    }
}

void consumer(int id, TaskPool& tp) {
    for(int i = 0; i < TASKS/THREADS; ++i){
        std::tuple<int,int,int> value = tp.execute_task();
        if (LOG) std::cout << "Consumer " << id << " executed task [" << std::get<1>(value) << "," << std::get<2>(value) << "] with value " << std::get<0>(value) << std::endl;
    }
}


int main(int argc, char* argv[]) {
    if (argc >= 2) {
        THREADS = std::stoi(argv[1]);
    }

    for (int i=0; i<MATRIX_DIM; i++) {
        matrix[i] = std::vector<int>(MATRIX_DIM);
        mutexes.push_back(std::vector<std::recursive_mutex>(MATRIX_DIM));
        for (int j = 0; j < MATRIX_DIM; j++)
            matrix[i][j] = rand() % 1000;
    }

    TaskPool tp(&matrix);

    std::vector<std::thread> consumers;

    auto t_start = std::chrono::high_resolution_clock::now();

    for (int i=0; i<THREADS; i++) {
        consumers.push_back(std::thread(consumer, i, std::ref(tp)));
    }

    std::thread p(producer, 0, std::ref(tp));

    p.join();
    for (int i=0; i<THREADS; i++)
        consumers[i].join();

    auto t_end = std::chrono::high_resolution_clock::now();
    double elapsed = std::chrono::duration<double, std::milli>(t_end-t_start).count()/1000.0;

    std::cout << "Done in " << elapsed << " s" << std::endl;
}

